import fs from "fs";
import url from "url";
import zlib from "zlib";
import stream from "stream";
import querystring from "querystring";
import multipart from "./multipart";
import createCookie from "./createCookie";

const stat = path=> new Promise((a,b)=>fs.stat(path, (e, d) => e?b(e):a(d) ))
/**
 * @param  {Function}	server		处理进程
 * @param  {Boolean}	cache		是否开启缓存
 * @param  {String[]}	path		请求路径（含域名）
 * @param  {Request}	req			客户端请求
 * @param  {Response}	res			访问权相应
 * @param  {Stting[]}	clientIp	客户端IP列表（考虑到代理）
 * @param  {Object}		cookie		cookie管理
 * @return {Boolean}				是否有效处理
 */
async function exec(server, cache, path, req, res, clientIp = [], cookie = {}) {
	if (!server) {
		return false;
	}
	//域名作为第一级路径
	const ret = await server(
		path, req.query, req.body,
		{
			method: req.method,
			path: req.path,
			domain: req.domainName,
			files: req.files,
			getCookie: cookie.get,
			setCookie: cookie.set,
			clearCookie: cookie.clear,
			getHeader: name => req.headers[name],
			setHeader: res.setHeader.bind(res),
			clientIp,
		}
	);
	if (!ret) {
		return false;
	}
	const {statusCode, download, transfer, filePath} = ret;
	let {mime, body, stream, gzip = true, length} = ret;

	let setCookie = cookie.header && cookie.header();
	setCookie = setCookie && setCookie.length && setCookie;

	// 302跳转
	if (transfer) {
		res.statusCode = 302;
		res.setHeader('Location', transfer);
		if (setCookie) {
			res.setHeader("Set-Cookie", setCookie);
		}
		res.end();
		return true;
	}

	let stats;
	//获取信息
	if (filePath) {
		try {
			stats = await stat(filePath);
		} catch(e) {
			return false;
		}
	}
	//使用文件缓存
	if (cache && filePath) {
		let time = req.headers['if-modified-since'];
		if (time && new Date(time) > stats.mtime) {
			res.statusCode = 304;
			if (setCookie) {
				res.setHeader("Set-Cookie", setCookie);
			}
			res.end();
			return true;
		}
	}
	//设置状态码与cookie
	res.statusCode = statusCode || 200;
	if (setCookie) {
		res.setHeader("Set-Cookie", setCookie);
	}

	//下载处理
	if (download && (filePath || stream)) {
		if (typeof download === "string") {
			res.setHeader('Content-Disposition', `attachment; filename="${encodeURIComponent(download)}"`);
		} else {
			res.setHeader('Content-Disposition', "attachment");
		}
		if (filePath) {
			// TODO: res.download(filePath);
			// return true;			
		}
	} else if (download) {
		res.setHeader('Content-Disposition', "attachment; filename=" + encodeURIComponent(download));
	}

	//文件转流
	if (filePath) {
		res.setHeader('Last-Modified', stats.mtime.toUTCString());
		stream = fs.createReadStream(filePath);
		if (stats.size < 1024 * 4) {
			length = stats.size;
			gzip = false;
		}
		if (!mime) {
			//TODO： 根据文件后缀获取流
		}
	}
	if (mime) {
		res.setHeader('Content-Type', mime);
	}

	//判断是否使用gzip压缩
	let zip = (req.headers["eccept-encoding"] || "").replace(/\s/g, "").toLowerCase().split(",");
	gzip = gzip && zip.indexOf("gzip") !== -1;

	//发送body
	if (stream) {
		if (gzip) {
			stream = stream.pipe(zlib.createGzip());
		} else if (length) {
			res.setHeader('Content-Length', length);
		}
		stream.pipe(res);
	} else if (body) {
		body = Buffer.from(body);
		if (body.length >= 1024 * 4 && gzip) {
			body = await new Promise(resolve => zlib.deflate(
				body,
				(err, buffer) => resolve(err ? '' : buffer)
			));
		}
		res.setHeader('Content-Length', body.length);
		res.end(body);
	} else {
		res.end();		
	}
	return true;
}

async function getStreamValue(stream) {
	return new Promise((resolve, reject) => {
		let allData = [];
		let size = 0;
		stream.on("error",reject);
		stream.on('data', function (data) {
			allData.push(data);
			size += data.length;
		});
		stream.on('end', function () {
			allData = Buffer.concat(allData, size);
			return resolve(allData);
		});
	});
}

async function getBody(req, res, saveFile, getUrl, getPath) {
	let charset, ctype = req.headers["content-type"];
	if (ctype) {
		[ctype, charset] = ctype.replace(/\s/g, "").split(";");
	}
	if (
		ctype === "application/json" ||
		ctype === "application/x-www-form-urlencoded" && (charset === "" || charset === "charset=UTF-8")
	) {
		try {
			let data = await getStreamValue(req);
			if (!data.length) {
				return;
			}
			data = data.toString();
			if (ctype === "application/json") {
				req.body = JSON.parse(data);
			} else {
				req.body = querystring.parse(data);
			}
		} catch(e) {}
		return ;
	}
	await multipart(req, saveFile, getUrl, getPath);
}

export default function server(staticServer, activeServer, saveFile, getUrl, getPath) {
	return async function server(req, res, next) {
		let beginTime = new Date;
		let {query, pathname: path} = url.parse(req.url, true);
		req.path = path;
		req.query = query;
		let host = req.headers.host;
		host = host ? host.replace(/(?:\:\d*)?$/g,"") : "";
		req.domainName = host;
		//域名作为第一级路径
		path = host + path;
		let method = req.method;
		let ua = req.headers['user-agent'];
		//静态路由，必须为GET请求且开启了静态服务器
		if (method === "GET" && await exec(staticServer, true, path, req, res)) {
			console.log(beginTime.toISOString(), (new Date - beginTime) / 1000, res.statusCode, 'static', method, path, ua);
			return;
		}
		//解析Cookie
		let cookie = createCookie(req.headers.cookie);
		// TODO: session
		// 解析body
		await getBody(req, res, saveFile, getUrl, getPath);
		
		// 获取客户端IP
		const clientIp = [req.socket.remoteAddress || ''];
		clientIp.push(req.headers['x-real-ip']);

		if (await exec(activeServer, false, path, req, res, clientIp, cookie)) {
			console.log(beginTime.toISOString(), (new Date - beginTime) / 1000, res.statusCode, 'active', method, path, ua);
			return ;
		}
		if (typeof next === "function") {
			console.log(beginTime.toISOString(), (new Date - beginTime) / 1000, 0, 'next', method, path, ua);
			return next();
		}
		console.log(beginTime.toISOString(), (new Date - beginTime) / 1000, 404, 'none', method, path, ua);
		res.statusCode = 404;
		res.end("404");
	};

}

